/* Atomix -- a little mind game about atoms and molecules.
 * Copyright (C) 1999 Jens Finke
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include "board.h"
#include <unistd.h>
#include "goal.h"
#include "undo.h"

#define                  ANIM_TIMEOUT     5  /* time in milliseconds between 
						two atom movements */

typedef struct _AnimData    AnimData;
typedef struct _MessageItems MessageItems;
typedef struct _LevelItems LevelItems;

struct _AnimData {
    gint              timeout_id;
    gint              counter;
    gint              dest_row;
    gint              dest_col;
    double            x_step;
    double            y_step;
};

struct _MessageItems
{
    GnomeCanvasGroup *messages;
    GnomeCanvasItem  *pause;
    GnomeCanvasItem  *game_over;
    GnomeCanvasItem  *new_game;
};

struct _LevelItems
{
    GnomeCanvasGroup *level;
    GnomeCanvasGroup *obstacles;
    GnomeCanvasGroup *moveables;
    GnomeCanvasGroup *decor;    
    GnomeCanvasGroup *background;    
};

/*=================================================================
 
  Global variables

  ---------------------------------------------------------------*/
static PlayField        *board = NULL;   /* the actual playfield */
static Goal             *goal = NULL;    /* the goal of this level */
static GnomeCanvasItem  *selected_item;  /* which canvas item currently be moved */
static AnimData         *anim_data;      /* holds the date for the atom animation */
static GSList           *board_canvas_items = NULL; /* a list of all used canvas items */
static MessageItems     *message_items;  /* references to the messages */
static LevelItems       *level_items;    /* references to the level groups */
static GnomeCanvasItem  *bgr_item;       /* reference to the background */
static gboolean         mouse_dragging;  /* whether the mouse is being 
					    dragged or not */
static GdkCursor        *hidden_cursor;  /* stores the transparent cursor */

/*=================================================================
 
  Declaration of internal functions

  ---------------------------------------------------------------*/
GnomeCanvasItem*
create_obstacle (double x, double y, AtomixObject *obj,
		 GnomeCanvasGroup* group);

GnomeCanvasItem*
create_moveable (double x, double y, AtomixObject *obj,
		 GnomeCanvasGroup* group);

GnomeCanvasItem*
create_item (GdkImlibImage* img, double x, double y,
	     GnomeCanvasGroup* group);

GnomeCanvasItem*
create_message(GnomeCanvasGroup* group, gchar* text);

GnomeCanvasGroup*
create_group(GnomeCanvasGroup* parent);

GnomeCanvasItem*
create_background(GnomeCanvasGroup* item);

void
create_hidden_cursor(void);

void
set_background_color(GnomeCanvasItem *item, gchar *outline, gchar *fill);

void
board_render(void);

gint
item_event (GnomeCanvasItem *item, GdkEvent *event, gpointer data);

void
move_item(GnomeCanvasItem* item, ItemDirection direc);

int
move_item_anim (void *data);

void
convert_to_playfield(gdouble x, gdouble y, guint *row, guint *col);

void
convert_to_canvas(guint row, guint col, gdouble *x, gdouble *y);


/*=================================================================
 
  Board creation, initialisation and clean up

  ---------------------------------------------------------------*/

void 
board_init (void) 
{
    /* Animation Data Setup */
    anim_data = g_malloc(sizeof(AnimData));
    anim_data->timeout_id = -1;
    anim_data->counter = 0;
    anim_data->dest_row = 0;
    anim_data->dest_col = 0;
    anim_data->x_step = 0.0;
    anim_data->y_step = 0.0;    

    /* Canvas setup */
    level_items = g_malloc(sizeof(LevelItems));

    /* init hidden cursor */
    create_hidden_cursor();

    /* init undo */
    undo_init();

    /* create the canvas groups for the level
     * (note: the first created group will be the first drawn group)
     */
    level_items->background = create_group(NULL);
    level_items->level = create_group(NULL);
    level_items->obstacles = create_group(level_items->level);
    level_items->decor = create_group(level_items->level);  /* not used yet */
    level_items->moveables = create_group(level_items->level);        

    /* create canvas group and items for the messages */
    message_items = g_malloc(sizeof(MessageItems));
    message_items->messages = create_group(NULL);
    message_items->pause = create_message(message_items->messages,
					  "Game Paused");
    message_items->game_over = create_message(message_items->messages,
					      "Game Over");
    message_items->new_game = create_message(message_items->messages,
					     "Press Ctrl+N for a new Game");

    bgr_item = create_background(level_items->background);
    gnome_canvas_item_show(message_items->new_game);
}

void
board_init_level (Level* l)
{
    anim_data->timeout_id = -1;
    anim_data->counter = 0;
    anim_data->dest_row = 0;
    anim_data->dest_col = 0;
    anim_data->x_step = 0.0;
    anim_data->y_step = 0.0;    

    selected_item = NULL;

    undo_free_all_moves();

    if(board) 
    {
	GnomeCanvas* canvas;
	gdouble x,y;

	board_clear();
	playfield_destroy(board);

	/* put level group to upper left corner */
	canvas = GNOME_CANVAS(lookup_widget(get_app(), "ca_matrix"));
	gnome_canvas_item_i2w(GNOME_CANVAS_ITEM(level_items->level), &x, &y);
	gnome_canvas_item_move(GNOME_CANVAS_ITEM(level_items->level), -x, -y);
    }
    
    board = playfield_copy(l->playfield);
    
    if(goal) 
    {
	goal_clear(goal);
	goal_destroy(goal);
    }
    goal = goal_new(l->goal);

    gnome_canvas_item_hide(message_items->new_game);

    board_render();
    goal_render(goal);
}

void
board_destroy()
{
    if(board) playfield_destroy(board);
    if(goal) goal_destroy(goal);
    if(anim_data) g_free(anim_data);
    
    undo_free_all_moves();

    if(level_items)
    {
	if(level_items->background)
	    gtk_object_destroy(GTK_OBJECT(level_items->background));

	if(level_items->obstacles)
	    gtk_object_destroy(GTK_OBJECT(level_items->obstacles));

	if(level_items->decor)
	    gtk_object_destroy(GTK_OBJECT(level_items->decor));

	if(level_items->moveables)
	    gtk_object_destroy(GTK_OBJECT(level_items->moveables));

	if(level_items->level)
	    gtk_object_destroy(GTK_OBJECT(level_items->level));

	g_free(level_items);
    }

    if(message_items)
    {
	if(message_items->pause)
	    gtk_object_destroy(GTK_OBJECT(message_items->pause));

	if(message_items->game_over)
	    gtk_object_destroy(GTK_OBJECT(message_items->game_over));

	if(message_items->new_game)
	    gtk_object_destroy(GTK_OBJECT(message_items->new_game));

	if(message_items->messages)
	    gtk_object_destroy(GTK_OBJECT(message_items->messages));

	g_free(message_items);
    }

    if(hidden_cursor)
    {
	gdk_cursor_destroy(hidden_cursor);
    }
}

/*=================================================================
 
  Board functions

  ---------------------------------------------------------------*/

void
board_render(void)
{
    GnomeCanvas *canvas;
    GnomeCanvasItem *item;
    gint row,col,width,height;
    gint tile_width, tile_height;
    gdouble x,y;
    AtomixObject* obj;
    AtomixType type;
    gdouble x1,x2,y1,y2;
    Theme *theme;

    theme = get_actual_theme();

    /* set background  colors*/
    set_background_color(bgr_item, theme->bgoutline, theme->bgcolor);
			   	
    /* create canvas items */
    for(row=0; row < board->n_rows; row++) 
    {
	for(col=0; col < board->n_cols; col++)
	{
	    obj = playfield_get_object(board, row, col);
	    type = obj->type;
	    switch(type)
	    {
	    case ATOMIX_MOVEABLE:
		convert_to_canvas(row, col, &x, &y);
		item = create_moveable(x, y, obj, level_items->moveables);
		break;

	    case ATOMIX_OBSTACLE:
		convert_to_canvas(row, col, &x, &y);
		item = create_obstacle(x, y, obj, level_items->obstacles);
		break;

	    case ATOMIX_NONE:
	    default:
	    }
	}
    }

    theme_tile_size(theme, &tile_width, &tile_height);

    /* center the whole thing */
    width = tile_width * board->n_cols;
    height = tile_height * board->n_rows;
    canvas = GNOME_CANVAS(lookup_widget(get_app(), "ca_matrix"));
    gnome_canvas_get_scroll_region(canvas, &x1, &y1, &x2, &y2);    
    x1 = (x2-x1-width)/2;
    y1 = (y2-y1-height)/2;    
    gnome_canvas_item_move(GNOME_CANVAS_ITEM(level_items->level), x1, y1);
}

gboolean
board_undo_move()
{
    if((anim_data->timeout_id == -1) &&
       (mouse_dragging == FALSE))
    {
	UndoMove *move = undo_get_last_move();

	if(move != NULL)
	{
	    gdouble x_src, y_src, x_dest, y_dest;
	    gint animstep;

	    playfield_swap_tiles(board, 
				 move->src_row, 
				 move->src_col, 
				 move->dest_row, 
				 move->dest_col);

	    convert_to_canvas(move->src_row, move->src_col, &x_src, &y_src);
	    convert_to_canvas(move->dest_row, move->dest_col, &x_dest, 
			      &y_dest);

	    animstep = get_actual_theme()->animstep;
	    if(move->src_col == move->dest_col)
	    {
		anim_data->counter = (gint)(fabs(y_dest - y_src) / animstep);
		anim_data->x_step = 0;
		anim_data->y_step = animstep;
		if(move->src_row < move->dest_row )
		{
		    anim_data->y_step = -(anim_data->y_step);
		}
	    }
	    else
	    {
		anim_data->counter = (gint)(fabs(x_dest - x_src) / animstep);
		anim_data->x_step = animstep;
		anim_data->y_step =  0;    
		if(move->src_col < move->dest_col)
		{
		    anim_data->x_step = -(anim_data->x_step);
		}
	    }
	    
	    anim_data->dest_row = move->src_row;
	    anim_data->dest_col = move->src_col;
	    selected_item = move->item;
	    
	    anim_data->timeout_id = gtk_timeout_add(ANIM_TIMEOUT, 
						    move_item_anim, 
						    anim_data);
	    g_free(move);
	}     	
	else return FALSE;
    } else return FALSE;

    return TRUE;
}

void
board_clear()
{
    set_background_color(bgr_item, "black", "cornsilk2");
    g_slist_foreach(board_canvas_items, destroy_item, NULL);
    g_slist_free(board_canvas_items);
    board_canvas_items = NULL;
}

void
board_print()
{
    g_print("Board:\n");
    playfield_print(board);
}

void board_view_message(gint msg_id)
{
    switch(msg_id)
    {
    case BOARD_MSG_NONE:
	gnome_canvas_item_hide(message_items->pause);
	gnome_canvas_item_hide(message_items->new_game);
	gnome_canvas_item_hide(message_items->game_over);
	break;
    case BOARD_MSG_GAME_PAUSED:
	gnome_canvas_item_show(message_items->pause);
	break;
    case BOARD_MSG_NEW_GAME:
	gnome_canvas_item_show(message_items->new_game);
	break;
    case BOARD_MSG_GAME_OVER:
	gnome_canvas_item_show(message_items->game_over);
	break;
    default:
    }
}

void board_hide_message(gint msg_id)
{
    switch(msg_id)
    {
    case BOARD_MSG_NONE:
	gnome_canvas_item_show(message_items->pause);
	gnome_canvas_item_show(message_items->new_game);
	gnome_canvas_item_show(message_items->game_over);
	break;
    case BOARD_MSG_GAME_PAUSED:
	gnome_canvas_item_hide(message_items->pause);
	break;
    case BOARD_MSG_NEW_GAME:
	gnome_canvas_item_hide(message_items->new_game);
	break;
    case BOARD_MSG_GAME_OVER:
	gnome_canvas_item_hide(message_items->game_over);
	break;
    default:
    }
}

void
board_hide(void)
{
    gnome_canvas_item_hide(GNOME_CANVAS_ITEM(level_items->level));
}

void
board_show(void)
{
    gnome_canvas_item_show(GNOME_CANVAS_ITEM(level_items->level));
}


void
board_show_normal_cursor(void)
{
    GtkWidget *canvas;
    GdkCursor *cursor;

    if(selected_item != NULL)
    {
	gnome_canvas_item_ungrab(selected_item, GDK_CURRENT_TIME);
    }

    canvas = GTK_WIDGET(lookup_widget(get_app(), "ca_matrix"));
    cursor = gdk_cursor_new(GDK_LEFT_PTR);
    gdk_window_set_cursor(gtk_widget_get_parent_window(canvas),
			  cursor);
    gdk_cursor_destroy(cursor);
}

/*=================================================================
 
  Board atom handling functions

  ---------------------------------------------------------------*/

gint
item_event (GnomeCanvasItem *item, GdkEvent *event, gpointer data)
{
    GtkWidget *canvas;
    GdkCursor *cursor;
    double x,y,diff_x, diff_y;

    static double selected_x, selected_y;
   
    switch (event->type) {
    case GDK_BUTTON_PRESS:
	
	if(anim_data->timeout_id!=-1) break; /* currently an object moved */

	if((item!=NULL) && (event->button.button==1)) 
	{
	    selected_x = event->button.x;
	    selected_y = event->button.y;
	    selected_item = item;
	    gnome_canvas_item_w2i (selected_item->parent, &selected_x,	    
				   &selected_y);

	    /* change cursor */	    
	    gnome_canvas_item_grab(item,
				   GDK_BUTTON_RELEASE_MASK | 
				   GDK_POINTER_MOTION_MASK,
				   hidden_cursor,
				   event->button.time);

	    mouse_dragging = TRUE;
	}
	break;

    case GDK_MOTION_NOTIFY:
	if(anim_data->timeout_id!=-1) break; /* currently an object moved */

	if (event->motion.state & GDK_BUTTON1_MASK) {

	    x = event->button.x;
	    y = event->button.y;
	    gnome_canvas_item_w2i (selected_item->parent, &x, &y);

	    diff_x = x - selected_x;
	    diff_y = y - selected_y;
	    
	    if((fabs(diff_x) > 8)|| (fabs(diff_y) > 8))
	    {
		/* ok we move the item */
		selected_x = x;
		selected_y = y;

		if(fabs(diff_x) > fabs(diff_y))
		{
		    /* move it horizontal */
		    if(diff_x < 0) 
		    {
			move_item(selected_item, LEFT);
		    }
		    else 
		    {
			move_item(selected_item, RIGHT);
		    }
		}
		else 
		{
		    /* move it vertical */
		    if(diff_y < 0)
		    {
			move_item(selected_item, UP);
		    }
		    else
		    {
			move_item(selected_item, DOWN);
		    }
		}
	    }
	}
	break;

    case GDK_BUTTON_RELEASE:
	board_show_normal_cursor();
	if(anim_data->timeout_id==-1) selected_item = NULL;
	mouse_dragging = FALSE;
	break;

    case GDK_ENTER_NOTIFY:
	if(!mouse_dragging)
	{
	    canvas = GTK_WIDGET(lookup_widget(get_app(), "ca_matrix"));
	    cursor = gdk_cursor_new(GDK_FLEUR);
	    gdk_window_set_cursor(gtk_widget_get_parent_window(canvas),
				  cursor);
	    gdk_cursor_destroy(cursor);
	}
	break;

    case GDK_LEAVE_NOTIFY:
	if(!mouse_dragging)
	{
	    board_show_normal_cursor();
	}
	break;

    default:
    }

    return FALSE;    
}

void
move_item(GnomeCanvasItem* item, ItemDirection direc)
{
    gdouble x1, y1, x2, y2;
    gdouble new_x1, new_y1;
    guint src_row, src_col, dest_row, dest_col;
    gint animstep;
    AtomixType type;
    UndoMove *move;
    
    gnome_canvas_item_get_bounds(item, &x1, &y1, &x2, &y2);
    convert_to_playfield(x1, y1, &src_row, &src_col);

    dest_row = src_row; dest_col = src_col;
    do
    {
	switch(direc)
	{
	case UP:
	    dest_row = dest_row - 1; break;
	case DOWN:
	    dest_row = dest_row + 1; break;
	case LEFT:
	    dest_col = dest_col - 1; break;
	case RIGHT:
	    dest_col = dest_col + 1; break;
	}
	type = playfield_get_object(board, dest_row, dest_col)->type;
    }
    while(type==ATOMIX_NONE);
    switch(direc)
    {
    case UP:
	dest_row++; break;
    case DOWN:
	    dest_row--; break;
    case LEFT:
	dest_col++; break;
    case RIGHT:
	dest_col--; break;
    }

    /* Remember this move for undo, but only if it is really a move
     * FIXME: Why is this sometimes called with src_row = dest_row and
     *        src_col = dest_col?
     */
    if((src_row != dest_row) || (src_col != dest_col))
    {
	move = undo_create_move(item, src_row, src_col, dest_row, dest_col);
	undo_add_move(move);
    }

    convert_to_canvas(dest_row, dest_col, &new_x1, &new_y1);
    playfield_swap_tiles(board, src_row, src_col, dest_row, dest_col);

    animstep = get_actual_theme()->animstep;
    if(direc == UP || direc == DOWN)
    {
	anim_data->counter = (gint)(fabs(new_y1 - y1) / animstep);
	anim_data->x_step = 0;
	anim_data->y_step = animstep;
	if(direc == UP)
	{
	    anim_data->y_step = -(anim_data->y_step);
	}
    }
    else
    {
	anim_data->counter = (gint)(fabs(new_x1 - x1) / animstep);
	anim_data->x_step = animstep;
	anim_data->y_step =  0;    
	if(direc == LEFT)
	{
	    anim_data->x_step = -(anim_data->x_step);
	}
    }

    anim_data->dest_row = dest_row;
    anim_data->dest_col = dest_col;
    anim_data->timeout_id = gtk_timeout_add(ANIM_TIMEOUT, move_item_anim, anim_data);
}

int
move_item_anim (void *data)
{
    AnimData* anim_data = (AnimData*)data;

    if(anim_data->counter > 0)
    {
	gnome_canvas_item_move(selected_item,
			       anim_data->x_step, anim_data->y_step);
	anim_data->counter--;
    }
    else
    {
	gtk_timeout_remove(anim_data->timeout_id);

	if(goal_reached(goal, board, anim_data->dest_row, anim_data->dest_col))
	{
	    game_level_finished();
	}

	anim_data->timeout_id = -1;
    }
    return 1;
}

/*=================================================================
 
  Internal creation functions

  ---------------------------------------------------------------*/

GnomeCanvasItem*
create_obstacle (double x, double y, AtomixObject *obj,
		 GnomeCanvasGroup* group)
{
  GdkImlibImage *img = NULL;
  GnomeCanvasItem *item = NULL;
  Theme* theme;
  gchar* file;

  theme = get_actual_theme();
  file = theme_get_object_image(theme, ATOMIX_OBSTACLE, obj->id);

  img = gdk_imlib_load_image(file);
  
  if(img) {
      item = create_item (img, x, y, group);
  } 
#ifdef DEBUG
  else {
      g_print("Wall image not found!\n");
  }
#endif    

  return item;
}

GnomeCanvasItem*
create_moveable (double x, double y, AtomixObject *obj, 
		 GnomeCanvasGroup* group)
{
  GdkImlibImage *img = NULL;
  GnomeCanvasItem *item = NULL;
  Theme* theme;
  gchar* file;

  theme = get_actual_theme();
  file = theme_get_object_image(theme, ATOMIX_MOVEABLE, obj->id);

  img = gdk_imlib_load_image(file);
  
  if(img) {
      item = create_item (img, x, y, group);
      gtk_signal_connect (GTK_OBJECT (item), "event",
			  (GtkSignalFunc) item_event,
			  NULL);      
  } 
#ifdef DEBUG
  else {
      g_print("Atom image not found!\n");
  }
#endif    

  return item;
}

GnomeCanvasItem*
create_item (GdkImlibImage* img, double x, double y,
	     GnomeCanvasGroup* group)
{
  GnomeCanvasItem *item;

  item = gnome_canvas_item_new(group,
			       gnome_canvas_image_get_type(),
			       "image", img,
			       "x", x,
			       "y", y,
			       "width", (double)img->rgb_width,
			       "height", (double)img->rgb_height,
			       "anchor", GTK_ANCHOR_NW,
			       NULL);                              
  gtk_signal_connect (GTK_OBJECT (item), "destroy",
		      (GtkSignalFunc) free_imlib_image, img);  

  board_canvas_items = g_slist_prepend(board_canvas_items, item);
  return item;
}

GnomeCanvasItem*
create_message(GnomeCanvasGroup* group, gchar* text)
{
    GnomeCanvasItem* item;
    GnomeCanvas* canvas;
    gdouble x, y, x1, y1, x2, y2;

    canvas = GNOME_CANVAS(lookup_widget(get_app(), "ca_matrix"));
    gnome_canvas_get_scroll_region(canvas, &x1, &y1, &x2, &y2);    
    x = x1 + (x2-x1)/2;
    y = y1 + (y2-y1)/2;    
    
    item = gnome_canvas_item_new(group,
				 gnome_canvas_text_get_type(),
				 "text", text,
				 "x", x,
				 "y", y,
				 "anchor", GTK_ANCHOR_CENTER,
				 "font","-*-helvetica-bold-r-*-*-14-*-*-*-*-*-*",
				 "justification", GTK_JUSTIFY_CENTER,
				 "fill_color", "black",
				 NULL);       

    gnome_canvas_item_hide(item);
    return item;
}

GnomeCanvasGroup*
create_group(GnomeCanvasGroup* parent)
{
    GnomeCanvasGroup* group;
    GnomeCanvas* canvas;

    canvas = GNOME_CANVAS(lookup_widget(get_app(), "ca_matrix"));
    if(parent == NULL)
    {
	parent = gnome_canvas_root(canvas);
    }
    group = GNOME_CANVAS_GROUP(gnome_canvas_item_new(
	parent,
	gnome_canvas_group_get_type(),
	"x", 0.0,
	"y", 0.0,
	NULL));			   

    return group;
}

GnomeCanvasItem*
create_background(GnomeCanvasGroup* group)
{
    GnomeCanvas* canvas;
    GnomeCanvasItem* item;
    gdouble x1,y1,x2,y2;

    canvas = GNOME_CANVAS(lookup_widget(get_app(), "ca_matrix"));
    gnome_canvas_get_scroll_region(canvas, &x1, &y1, &x2, &y2);

    item = gnome_canvas_item_new(group,
				gnome_canvas_rect_get_type(),
				"x1", x1,
				"y1", y1,
				"x2", x2,
				"y2", y2,
				"outline_color", "black",
				"fill_color", "cornsilk2",
				"width_pixels", 2,
				NULL);
    return item;
}

void
create_hidden_cursor(void)
{
    GdkColor white = {0, 0xffff, 0xffff, 0xffff};
    GdkColor black = {0, 0x0000, 0x0000, 0x0000};
    GdkBitmap *bitmap;
    GdkBitmap *mask;
    gchar data[] = { 0 };

    bitmap = gdk_bitmap_create_from_data(NULL, data, 1, 1);
    mask = gdk_bitmap_create_from_data(NULL, data, 1, 1);
    
    hidden_cursor = gdk_cursor_new_from_pixmap(bitmap, mask, 
					       &white, &black, 1, 1);    
}


/*=================================================================
 
  Internal helper functions

  ---------------------------------------------------------------*/

void
set_background_color(GnomeCanvasItem *item, gchar *outline, gchar *fill)
{
    if(outline && fill && item)
    {
	GdkColor gdk_outline;
	GdkColor gdk_fill;

	if(gdk_color_parse(outline, &gdk_outline) &&
	   gdk_color_parse(fill, &gdk_fill) &&
	   gdk_colormap_alloc_color(gdk_colormap_get_system(),
				    &gdk_outline, FALSE, TRUE) &&
	   gdk_colormap_alloc_color(gdk_colormap_get_system(),
				    &gdk_fill, FALSE, TRUE)
	    )
	{	    
	    gnome_canvas_item_set(item, "outline_color_gdk", &gdk_outline,
				  "fill_color_gdk", &gdk_fill, NULL);
	}
    }
}

void
convert_to_playfield(gdouble x, gdouble y, guint *row, guint *col) 
{
    guint int_y, int_x;
    gint tile_width, tile_height;

    theme_tile_size(get_actual_theme(), &tile_width, &tile_height);

    int_y = (guint) ceil(y);
    *row = (int_y / tile_height);
    
    int_x = (guint) ceil(x);
    *col = (int_x / tile_width);
}

void
convert_to_canvas(guint row, guint col, gdouble *x, gdouble *y) 
{
    gint tile_width, tile_height;

    theme_tile_size(get_actual_theme(), &tile_width, &tile_height);

    *x = col * tile_width;
    *y = row * tile_height;
}
