/* Atomix -- a little mind game about atoms and molecules.
 * Copyright (C) 1999 Jens Finke
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <sys/types.h>
#include <sys/stat.h>
#include <stdlib.h>
#include <fcntl.h>
#include <dirent.h>
#include <parser.h> /* from libxml */
#include <tree.h>   /* from libxml */

#include "level.h"

static const GScannerConfig	level_scanner_config =
{
  (
   " \t\n"
   )			/* cset_skip_characters */,
  (
   G_CSET_a_2_z
   "_"
   G_CSET_A_2_Z
   )			/* cset_identifier_first */,
  (
   G_CSET_a_2_z
   "_-0123456789"
   G_CSET_A_2_Z
   )			/* cset_identifier_nth */,
  ( "#\n" )		/* cpair_comment_single */,
  
  FALSE			/* case_sensitive */,
  
  TRUE			/* skip_comment_multi */,
  TRUE			/* skip_comment_single */,
  TRUE			/* scan_comment_multi */,
  TRUE			/* scan_identifier */,
  FALSE			/* scan_identifier_1char */,
  FALSE			/* scan_identifier_NULL */,
  TRUE			/* scan_symbols */,
  TRUE			/* scan_binary */,
  TRUE			/* scan_octal */,
  TRUE			/* scan_float */,
  TRUE			/* scan_hex */,
  TRUE			/* scan_hex_dollar */,
  TRUE			/* scan_string_sq */,
  TRUE			/* scan_string_dq */,
  TRUE			/* numbers_2_int */,
  FALSE			/* int_2_float */,
  FALSE			/* identifier_2_string */,
  TRUE			/* char_2_token */,
  TRUE			/* symbol_2_token */,
  FALSE			/* scope_0_fallback */,
};

static const struct
{
  gchar *name;
  guint token;
} symbols[] = {
  { "level", LEVEL_TOKEN_LEVEL },
  { "time", LEVEL_TOKEN_TIME },
  { "theme", LEVEL_TOKEN_THEME },
  { "next", LEVEL_TOKEN_NEXT },
  { "playfield", LEVEL_TOKEN_PLAYFIELD },
  { "goal", LEVEL_TOKEN_GOAL }
};

static const guint n_symbols = sizeof (symbols) / sizeof (symbols[0]);

/* level parse functions */
gchar*
level_get_level_name(gchar* filename);

gchar*
level_get_level_name_xml(gchar* filename);

gboolean
level_parse_time(GScanner* scanner, Level* level);

gboolean
level_parse_next(GScanner* scanner, Level* level);

gboolean
level_parse_theme(GScanner* scanner, Level* level);

gboolean
level_parse_goal(GScanner* scanner, Level* level);

gboolean
level_parse_playfield(GScanner* scanner, Level* level);

gboolean
level_parse_raw_playfield(GScanner* scanner, PlayField* pf);


/* hash table helper functions */
void
level_print_hash_table(gpointer key, gpointer value, gpointer user_data);

void
level_destroy_hash_table_item(gpointer key, gpointer value, gpointer user_data);

/*************************************************************
 *  Implementation
 */

Level* 
level_new(void)
{
    Level* p_level;
    p_level = g_malloc(sizeof(Level));    

    p_level->name = NULL;
    p_level->next = NULL;
    p_level->theme_name = NULL;
    p_level->time = 0;
    p_level->playfield = NULL;
    p_level->goal = NULL;

    return p_level;
}

void
level_destroy(Level* l) 
{
    if(l!=NULL)
    {
	if(l->playfield) 
	{
		playfield_destroy(l->playfield);
	}
	if(l->goal)
	{
		playfield_destroy(l->goal);
	}
	if(l->name)
	{
		g_free(l->name);
	}
	if(l->next)
	{
		g_free(l->next);
	}
	if(l->theme_name)
	{
		g_free(l->theme_name);
	}    
	g_free(l);
    }
}

void
level_create_hash_table(void)
{
    gint i, n_dirs;
    gchar* users_home;
    gchar* search_dirs[3];
    struct dirent *dent = NULL;
    DIR* dir;

    available_levels = g_hash_table_new((GHashFunc) g_str_hash, 
					(GCompareFunc)g_str_equal);
    
    /* define directorys where to search for levels */
    users_home = g_get_home_dir();
    
    n_dirs = 3;
    search_dirs[0] = g_strdup("../levels");
    search_dirs[1] = g_strjoin("/", users_home, ".atomix/levels", NULL);
    search_dirs[2] = g_strjoin("/", DATADIR, "atomix/levels", NULL);
    
    /* search the dirs */
    for(i = 0; (i < n_dirs); i++)
    {
	dir = opendir(search_dirs[i]);
	if(dir)
	{
	    gchar* filename;
	    gchar *filepath;
	    gchar* levelname = NULL;
	    while((dent = readdir(dir)) != NULL)
	    {
		if((g_strcasecmp(".", dent->d_name)!=0) &&
		   (g_strcasecmp("..", dent->d_name)!=0))
		{
		    filepath = g_strjoin("/", search_dirs[i], dent->d_name, NULL);
		    filename = g_strdup(dent->d_name);
		    g_print("Check File: %s\n",filename);
		    g_strreverse(filename);
		    
		    if(g_strncasecmp(filename, "lmx.", 4) == 0)
		    {
			levelname = level_get_level_name_xml(filepath);
			g_print(" -> is xml level: %s\n", levelname);
		    }
		    else if(g_strncasecmp(filename, "vel.", 4) == 0)
		    {
			/*levelname = level_get_level_name(filepath);
			  g_print(" -> is lev level: %s\n", levelname);*/
			g_print(" -> ignoring \n");
		    }
		    else
		    {
			g_print(" -> ignoring \n");
		    }
		    
		    if(levelname)
		    {
			g_hash_table_insert(available_levels, 
					    g_strdup(levelname),
					    g_strdup(filepath));
		    }
		    g_free(filename);
		    g_free(filepath);
		    if(levelname) g_free(levelname);
		    levelname = NULL;
		}
	    }
	}
#ifdef DEBUG
	else
	{
	    //    g_print("Opendir error: %s\n", search_dirs[i]);
	}
#endif

	//g_free(dent);
	closedir(dir);
    }
    
    for(i = 0; i < n_dirs; i++) g_free(search_dirs[i]);

#ifdef DEBUG
/*    g_print("Scanning for Levels, found: \n");
      g_hash_table_foreach(available_levels, (GHFunc) level_print_hash_table, NULL); */
#endif
}

void
level_print_hash_table(gpointer key, gpointer value, gpointer user_data)
{
    g_print("Key: %s, Value: %s\n",(gchar*) key, (gchar*) value);
}

void
level_destroy_hash_table(void)
{
    if(available_levels)
    {
	g_hash_table_foreach(available_levels, 
			     (GHFunc) level_destroy_hash_table_item, NULL);
	g_hash_table_destroy(available_levels);
    }
}

void
level_destroy_hash_table_item(gpointer key, gpointer value, gpointer user_data)
{
    g_free(key);
    g_free(value);
}

gchar*
level_get_first_level(void)
{
    /* FIXME: need implementation here */
    return "Water";
}

Level*
level_load(gchar* name, LoadResult* result)
{
    Level* level = NULL;
    gint fd;
    gchar* path = NULL;

    *result = OK;

    path = (gchar*) g_hash_table_lookup(available_levels, name);

    fd = open (path ,O_RDONLY);

    if(fd==-1) 
    {
	g_print("Level not found!\n");
	*result = NOT_FOUND;
    }

    else
    {
	level = level_new();
	if(!level_parse_level(fd, level))
	{
	    g_print("This is not a valid AtomiX level!\n");
	    *result = NOT_A_VALID_FILE;
	    level_destroy(level);
	    level = NULL;
	}
    }

    close(fd);
    return level;
}

Level* 
level_load_xml(gchar* name)
{
    xmlDocPtr doc;
    xmlNodePtr node;
    Level *level;
    gchar *prop_value;
    gchar *path;

    level = level_new();

    /* read file */
    path = (gchar*) g_hash_table_lookup(available_levels, name);
    doc = xmlParseFile(path); 
       
    if(doc != NULL)
    {
	node = doc->root;
	
	while(node!=NULL)
	{
	    if(g_strcasecmp(node->name,"LEVEL")==0)
	    {
		/* handle level node */
		prop_value = xmlGetProp(node, "name");
		level->name = g_strdup(prop_value);
		g_print("Level Name: %s\n", level->name);
		
		node = node->childs;
	    }
	    else 
	    {
		if(g_strcasecmp(node->name,"TIME")==0)
		{
		    /* handle time node */
		    prop_value = xmlGetProp(node, "secs");
		    level->time = atoi(prop_value);
		    g_print("Time: %i\n", level->time);
		}
		
		else if(g_strcasecmp(node->name,"THEME")==0)
		{
		    /* handle theme node */
		    prop_value = xmlGetProp(node, "name");
		    level->theme_name = g_strdup(prop_value);
		    g_print("Theme Name: %s\n", level->theme_name);
		}
		
		else if(g_strcasecmp(node->name,"NEXT")==0)
		{
		    /* handle next node */
		    prop_value = xmlGetProp(node, "level");
		    level->next = g_strdup(prop_value);
		    g_print("Next Level: %s\n", level->next);
		    
		}
		
		else if(g_strcasecmp(node->name,"PLAYFIELD")==0)
		{
		    /* handle playfield node */
		    level->playfield = playfield_load_xml(node);
		    playfield_print(level->playfield);
		}
		
		else if(g_strcasecmp(node->name,"GOAL")==0)
		{
		    /* handle goal node */
		    level->goal =  playfield_load_xml(node);
		    playfield_print(level->goal);
		}

		else
		{
		    g_print("Unknown TAG, ignoring <%s>\n",node->name);
		}
		
		node = node->next;
	    }	    
	}
	xmlFreeDoc(doc);	    
    }
    else
    {
	return NULL;
    }
    
    return level;
}

void
level_save_xml(Level* level, gchar* filename)
{
    xmlDocPtr doc;
    xmlAttrPtr attr;
    xmlNodePtr level_node;
    xmlNodePtr playfield_node;
    xmlNodePtr goal_node;
    xmlNodePtr child_node;
    gchar *str_buffer; 
    gint length;

    str_buffer = g_malloc(5*sizeof(gchar));

    /* create xml doc */    
    doc = xmlNewDoc("1.0");
    xmlSetDocCompressMode(doc, 9);

    /* level name */
    level_node = xmlNewDocNode(doc, NULL, "LEVEL", NULL);
    doc->root = level_node;
    attr = xmlSetProp(level_node, "name", g_strdup(level->name));

    /* time */
    child_node = xmlNewChild(level_node, NULL, "TIME", NULL);
    length = g_snprintf(str_buffer, 5, "%i", level->time);
    attr = xmlSetProp(child_node, "secs", g_strdup(str_buffer));
    
    /* theme */
    child_node = xmlNewChild(level_node, NULL, "THEME", NULL);
    attr = xmlSetProp(child_node, "name", g_strdup(level->theme_name));

    /* next */
    child_node = xmlNewChild(level_node, NULL, "NEXT", NULL);
    attr = xmlSetProp(child_node, "level", g_strdup(level->next));

    /* Playfield */
    playfield_node = xmlNewChild(level_node, NULL, "PLAYFIELD", NULL);
    playfield_save_xml(level->playfield, playfield_node);

    /* Goal */
    goal_node = xmlNewChild(level_node, NULL, "GOAL", NULL);
    playfield_save_xml(level->goal, goal_node);
    
    xmlSaveFile(filename, doc);

    xmlFreeDoc(doc);
    g_free(str_buffer);
}

gchar*
level_get_level_name_xml(gchar* filename)
{
    xmlDocPtr doc;
    xmlNodePtr node;
    gchar* name = NULL;
    gchar *prop_value;

    /* read file */
    doc = xmlParseFile(filename); 
       
    if(doc != NULL)
    {
	node = doc->root;
	
	if((node!=NULL) &&
	   (g_strcasecmp(node->name,"LEVEL")==0))
	{
	    /* handle level node */
	    prop_value = xmlGetProp(node, "name");
	    name = g_strdup(prop_value);
	}
    }

    xmlFreeDoc(doc);
    
    return name;
}


gchar*
level_get_level_name(gchar* filename)
{
    GScanner *scanner;
    gint fd;
    gint i;
    guint token;
    gchar* name  = NULL;

    scanner = g_scanner_new ((GScannerConfig *) &level_scanner_config);
    fd = open(filename, O_RDONLY);
    
    if(fd==-1)
    {
	g_print("level_get_level_name: file not found!\n");
	g_scanner_destroy(scanner);
	return NULL;
    }
    else 
    {

	g_scanner_freeze_symbol_table (scanner);
	for (i = 0; i < n_symbols; i++)
	{
	    g_scanner_add_symbol (scanner, symbols[i].name, 
				  GINT_TO_POINTER (symbols[i].token));
	}
	g_scanner_thaw_symbol_table (scanner);

	g_scanner_input_file(scanner, fd);

	if (g_scanner_peek_next_token(scanner) != G_TOKEN_EOF)
	{
	    token = g_scanner_get_next_token(scanner);
	    if((token == LEVEL_TOKEN_LEVEL) &&
	       (g_scanner_peek_next_token(scanner) != G_TOKEN_EOF))
	    {
		token = g_scanner_get_next_token(scanner);
		
		if(token == G_TOKEN_STRING)
		{
		    name = g_strdup(scanner->value.v_string); 
		} 
		 else g_print("No Identifier\n");
	    } 
	    else g_print("No Level token\n"); 
	}
	 else g_print("EOF\n"); 
    }
    g_scanner_destroy(scanner);
    close(fd);

    return name;
}
    
    
gboolean
level_parse_level(guint fd, Level* level)
{
    GScanner *scanner;
    int i;
    LevelFSM state = LEVEL_NONE_READ;
    guint token;
    gboolean parse_error = FALSE;

    scanner = g_scanner_new ((GScannerConfig *) &level_scanner_config);
    g_scanner_input_file(scanner, fd);
    
    g_scanner_freeze_symbol_table (scanner);
    for (i = 0; i < n_symbols; i++)
    {
	g_scanner_add_symbol (scanner, symbols[i].name, 
			      GINT_TO_POINTER (symbols[i].token));
    }
    g_scanner_thaw_symbol_table (scanner);

    while((g_scanner_peek_next_token(scanner) != G_TOKEN_EOF) &&
	  !parse_error && (state != LEVEL_CLOSED_BRACE))
    {
	token = g_scanner_get_next_token(scanner);
	switch(state)
	{
	case LEVEL_NONE_READ:
	    if(token == LEVEL_TOKEN_LEVEL)
	    {
		state = LEVEL_LEVEL_READ;
	    }
	    else
	    {
		parse_error = TRUE;
	    }
	    break;

	case LEVEL_LEVEL_READ:
	    if(token == G_TOKEN_STRING)
	    {
		level->name = g_strdup(scanner->value.v_string);
		state = LEVEL_LEVEL_IDENTIFIER;
	    }
	    else
	    {
		parse_error = TRUE;
	    }
	    break;

	case LEVEL_LEVEL_IDENTIFIER:
	    if(token == G_TOKEN_LEFT_CURLY)
	    {
		state = LEVEL_CONTENT;
	    }
	    else
	    {
		parse_error = TRUE;
	    }
	    break;

	case LEVEL_CONTENT:
	    switch(token)
	    {
	    case LEVEL_TOKEN_TIME:
		parse_error = level_parse_time(scanner, level);
		break;

	    case LEVEL_TOKEN_NEXT:
		parse_error = level_parse_next(scanner, level);
		break;

	    case LEVEL_TOKEN_THEME:
		parse_error = level_parse_theme(scanner, level);
		break;

	    case LEVEL_TOKEN_PLAYFIELD:
		parse_error = level_parse_playfield(scanner, level);
		break;

	    case LEVEL_TOKEN_GOAL:
		parse_error = level_parse_goal(scanner, level);
		break;

	    case G_TOKEN_RIGHT_CURLY:
		state = LEVEL_CLOSED_BRACE;
		break;

	    default:
		parse_error = TRUE;		
	    }	    	   
	    break;

	default:
	    parse_error = TRUE;
	}
    }

    g_scanner_destroy(scanner);

    if(state!=LEVEL_CLOSED_BRACE) parse_error = TRUE;

    return !parse_error;
}

gboolean
level_parse_time(GScanner* scanner, Level* level)
{
    guint token;
    LevelFSM state = LEVEL_NONE_READ;
    gboolean parse_error = FALSE;

    while((g_scanner_peek_next_token(scanner) != G_TOKEN_EOF) &&
	  !parse_error && (state != LEVEL_VALUE_READ))
    {
	token = g_scanner_get_next_token(scanner);
	switch(state)
	{
	case LEVEL_NONE_READ:
	    if(token == G_TOKEN_EQUAL_SIGN)
	    {
		state = LEVEL_EQUAL_READ;
	    }
	    else
	    {
		parse_error = TRUE;
	    }
	    break;

	case LEVEL_EQUAL_READ:
	    if(token == G_TOKEN_INT)
	    {
		level->time = scanner->value.v_int;
		state = LEVEL_VALUE_READ;
	    }
	    else
	    {
		parse_error = TRUE;
	    }
	    break;

	default:
	    parse_error = TRUE;
	    
	}
    }

    if(state!=LEVEL_VALUE_READ) parse_error = TRUE;

    return parse_error;
}

gboolean
level_parse_next(GScanner* scanner, Level* level)
{
    guint token;
    LevelFSM state = LEVEL_NONE_READ;
    gboolean parse_error = FALSE;

    while((g_scanner_peek_next_token(scanner) != G_TOKEN_EOF) &&
	  !parse_error && (state != LEVEL_VALUE_READ))
    {
	token = g_scanner_get_next_token(scanner);
	switch(state)
	{
	case LEVEL_NONE_READ:
	    if(token == G_TOKEN_EQUAL_SIGN)
	    {
		state = LEVEL_EQUAL_READ;
	    }
	    else
	    {
		parse_error = TRUE;
	    }
	    break;

	case LEVEL_EQUAL_READ:
	    if(token == G_TOKEN_STRING)
	    {
		level->next = g_strdup(scanner->value.v_string);
		state = LEVEL_VALUE_READ;
	    }
	    else
	    {
		parse_error = TRUE;
	    }
	    break;

	default:
	    parse_error = TRUE;
	    
	}
    }

    if(state!=LEVEL_VALUE_READ) parse_error = TRUE;

    return parse_error;
}

gboolean
level_parse_theme(GScanner* scanner, Level* level)
{
    guint token;
    LevelFSM state = LEVEL_NONE_READ;
    gboolean parse_error = FALSE;

    while((g_scanner_peek_next_token(scanner) != G_TOKEN_EOF) &&
	  !parse_error && (state != LEVEL_VALUE_READ))
    {
	token = g_scanner_get_next_token(scanner);
	switch(state)
	{
	case LEVEL_NONE_READ:
	    if(token == G_TOKEN_EQUAL_SIGN)
	    {
		state = LEVEL_EQUAL_READ;
	    }
	    else
	    {
		parse_error = TRUE;
	    }
	    break;

	case LEVEL_EQUAL_READ:
	    if(token == G_TOKEN_IDENTIFIER)
	    {
		level->theme_name = g_strdup(scanner->value.v_string);
		state = LEVEL_VALUE_READ;
	    }
	    else
	    {
		parse_error = TRUE;
	    }
	    break;

	default:
	    parse_error = TRUE;
	    
	}
    }

    if(state!=LEVEL_VALUE_READ) parse_error = TRUE;

    return parse_error;
}

gboolean
level_parse_goal(GScanner* scanner, Level* level)
{
    gboolean parse_error = FALSE;
    PlayField* pf = NULL;
    
    pf = playfield_new();
    g_assert(pf);
    parse_error = level_parse_raw_playfield(scanner, pf);
    
    if(!parse_error)
    {
	level->goal = pf;
    }
    else if(pf)
    {
	playfield_destroy(pf);
	level->goal = NULL;
    }
    
    return parse_error;
}

gboolean
level_parse_playfield(GScanner* scanner, Level* level)
{
    gboolean parse_error = FALSE;
    PlayField* pf;

    pf = playfield_new();
    g_assert(pf);
    parse_error = level_parse_raw_playfield(scanner, pf);

    g_assert(pf);
    if(!parse_error)
    {
	level->playfield = pf;
    }
    else if(pf)
    {
	playfield_destroy(pf);
	level->playfield = NULL;
    }
    
    return parse_error;
}

gboolean
level_parse_raw_playfield(GScanner* scanner, PlayField* pf)
{
    guint token;
    LevelFSM state = LEVEL_NONE_READ;
    gboolean parse_error = FALSE;
    guint row = -1;
    guint col = -1;
    AtomixObject* obj;
    AtomixType type = ATOMIX_NONE;
    guint id;

    while((g_scanner_peek_next_token(scanner) != G_TOKEN_EOF) &&
	  !parse_error && (state != LEVEL_CLOSED_BRACE))
    {
	token = g_scanner_get_next_token(scanner);
	switch(state)
	{
	case LEVEL_NONE_READ:
	    //g_print("NONE_READ\n");
	    if(token == G_TOKEN_EQUAL_SIGN)
	    {
		state = LEVEL_EQUAL_READ;		
	    }
	    else
	    {
		parse_error = TRUE;
	    }
	    break;

	case LEVEL_EQUAL_READ:
	    //g_print("EQUAL_READ\n");
	    if(token == G_TOKEN_LEFT_CURLY)
	    {
		state = LEVEL_OPEN_BRACE;
	    }
	    else
	    {
		parse_error = TRUE;
	    }
	    break;

	case LEVEL_OPEN_BRACE:
	    //g_print("OPEN_BRACE\n");
	    if(token == G_TOKEN_LEFT_CURLY)
	    {
		state = LEVEL_READ_ROW;
		playfield_add_row(pf);
		row++;
	    }
	    else if(token == G_TOKEN_RIGHT_CURLY)
	    {
		state = LEVEL_CLOSED_BRACE;
	    }
	    else
	    {
		parse_error = TRUE;
	    }
	    break;

	case LEVEL_READ_ROW:
	    //g_print("READ_ROW\n");
	    if(token == G_TOKEN_INT)
	    {
		type = (AtomixType)scanner->value.v_int;
		//g_print("%x|", type);
		state = LEVEL_TYPE_READ;
	    }
	    else
	    {
		parse_error = TRUE;
	    }
	    break;

	case LEVEL_TYPE_READ:
	    //g_print("TYPE_READ\n");
	    if(token == G_TOKEN_COMMA)
	    {
		state = LEVEL_COMMA_READ;
	    }
	    else
	    {
		parse_error = TRUE;
	    }
	    break;
	    
	case LEVEL_COMMA_READ:
	    //g_print("COMMA_READ\n");
	    if(token == G_TOKEN_INT)
	    {
		id = (guint)scanner->value.v_int;
		if(row==0)
		{
		    playfield_add_column(pf);
		}
		col++;
		obj = atomix_object_new_args(type, id);
		playfield_set_object(pf, obj, row, col);
		//g_print("%x ",(int)id);
		state = LEVEL_COL_READ;
	    }
	    else
	    {
		parse_error = TRUE;
	    }
	    break;

	case LEVEL_COL_READ:
	    //g_print("COL_READ\n");
	    if(token == G_TOKEN_RIGHT_CURLY)
	    {
		col = -1;
		//g_print("\n");
		state = LEVEL_ROW_READ;				
	    }
	    else if(token == G_TOKEN_COMMA)
	    {
		state = LEVEL_READ_ROW;
	    }
	    else
	    {
		parse_error = TRUE;
	    }	    
	    break;

	case LEVEL_ROW_READ:
	    //g_print("ROW_READ\n");
	    if(token == G_TOKEN_RIGHT_CURLY)
	    {	
		state = LEVEL_CLOSED_BRACE;
	    }
	    else if(token == G_TOKEN_COMMA)
	    {
		state = LEVEL_OPEN_BRACE;
	    }
	    else
	    {
		parse_error = TRUE;
	    }	    
	    break;

	default:
	    //g_print("DEFAULT\n");
	    parse_error = TRUE;
	    
	}
    }
    
    if(state!=LEVEL_CLOSED_BRACE) parse_error = TRUE;

    return parse_error;
}

