/* Atomix -- a little mind game about atoms and molecules.
 * Copyright (C) 1999 Jens Finke
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <sys/types.h>
#include <sys/stat.h>
#include <dirent.h>
#include <fcntl.h>
#include <parser.h> /* from libxml */
#include <tree.h>   /* from libxml */


#include "theme.h"

static const GScannerConfig	theme_scanner_config =
{
  (
   " \t\n"
   )			/* cset_skip_characters */,
  (
   G_CSET_a_2_z
   "_"
   G_CSET_A_2_Z
   )			/* cset_identifier_first */,
  (
   G_CSET_a_2_z
   "_-0123456789"
   G_CSET_A_2_Z
   )			/* cset_identifier_nth */,
  ( "#\n" )		/* cpair_comment_single */,
  
  FALSE			/* case_sensitive */,
  
  TRUE			/* skip_comment_multi */,
  TRUE			/* skip_comment_single */,
  TRUE			/* scan_comment_multi */,
  TRUE			/* scan_identifier */,
  FALSE			/* scan_identifier_1char */,
  FALSE			/* scan_identifier_NULL */,
  TRUE			/* scan_symbols */,
  TRUE			/* scan_binary */,
  TRUE			/* scan_octal */,
  TRUE			/* scan_float */,
  TRUE			/* scan_hex */,
  TRUE			/* scan_hex_dollar */,
  TRUE			/* scan_string_sq */,
  TRUE			/* scan_string_dq */,
  TRUE			/* numbers_2_int */,
  FALSE			/* int_2_float */,
  FALSE			/* identifier_2_string */,
  TRUE			/* char_2_token */,
  TRUE			/* symbol_2_token */,
  FALSE			/* scope_0_fallback */,
};

static const struct
{
  gchar *name;
  guint token;
} symbols[] = {
  { "theme", THEME_TOKEN_THEME },
  { "obstacle", THEME_TOKEN_OBSTACLE },
  { "moveable", THEME_TOKEN_MOVEABLE },
  { "decor", THEME_TOKEN_DECOR }
};

static const guint n_symbols = sizeof (symbols) / sizeof (symbols[0]);

void 
theme_fill_img_array(Theme *theme, xmlNodePtr node, AtomixType type);

gboolean
theme_parse_moveable(GScanner* scanner, Theme* theme);

gboolean
theme_parse_obstacle(GScanner* scanner, Theme* theme);

gboolean
theme_parse_decor(GScanner* scanner, Theme* theme);

gboolean
theme_parse_theme(guint fd, Theme* theme);

gboolean
theme_parse_image_list(GScanner* scanner, Theme* theme, AtomixType type);

gchar*
theme_get_theme_name(gchar* filename);

gchar*
theme_get_theme_name_xml(gchar *filename);

void
theme_print_hash_table(gpointer key, gpointer value, gpointer user_data);

void
theme_destroy_hash_table_item(gpointer key, gpointer value, gpointer user_data);

/**
 * Creates a new Theme object. 
 * Returns a pointer to the new created object.
 */
Theme*
theme_new()
{
    Theme* theme;
    theme = g_malloc(sizeof(Theme));

    theme->name = NULL;
    theme->path = NULL;
    theme->obstacles = g_ptr_array_new();
    theme->moveables = g_ptr_array_new();
    theme->decor = NULL;
    theme->tile_width = 0;
    theme->tile_height = 0;
    theme->animstep = 4;
    theme->bgcolor = NULL;
    theme->bgoutline = NULL;

    return theme;
}

/**
 * Deletes a Theme object. 
 * theme - to be destroyed theme
 */
void
theme_destroy(Theme* theme)
{
    if(theme) 
    {	
	if(theme->name)
	{
	    g_free(theme->name);
	}
	if(theme->path)
	{
	    g_free(theme->path);
	}
	if(theme->obstacles) 
	{
	    g_ptr_array_free(theme->obstacles, TRUE);
	}
	if(theme->moveables)
	{
	    g_ptr_array_free(theme->moveables, TRUE);
	}
	if(theme->decor)
	{
	    g_ptr_array_free(theme->decor, TRUE);
	}
	if(theme->bgcolor)
	{
	    g_free(theme->bgcolor);	  
	}
	if(theme->bgoutline)
	{
	    g_free(theme->bgoutline);	  
	}
	g_free(theme);
    }
}

/**
 * Adds an image name to a specific kind of object.
 * theme - the theme to which the image should be added
 * type - the type of object 
 *        (possibilites: ATOMIX_MOVEABLE, ATOMIX_OBSTACLE, ATOMIX_DECOR (not supported yet)
 * file - the name of the image file
 */
void
theme_add_object_image(Theme* theme, AtomixType type, gchar* file)
{
    if(theme!=NULL) 
    {
	switch(type)
	{
	case ATOMIX_OBSTACLE:
	    if(theme->obstacles) 
	    {
		g_ptr_array_add(theme->obstacles, (gpointer) file);
	    }
	    break;
	    
	case ATOMIX_MOVEABLE:
	    if(theme->moveables)
	    {
		g_ptr_array_add(theme->moveables, (gpointer) file);
	    }
	    break;

	case ATOMIX_DECOR:
	    if(theme->decor)
	    {
		g_ptr_array_add(theme->decor, (gpointer) file);
	    }
	    break;
	    
	default:
	}

	if((theme->tile_width == 0) || (theme->tile_height == 0))
	{
	    GdkImlibImage *img = NULL;
	    gchar *full_path;

	    full_path = g_strjoin("/", theme->path, file, NULL);
	    img = gdk_imlib_load_image(full_path);
	    theme->tile_width = img->rgb_width;
	    theme->tile_height = img->rgb_height;
	    
	    g_print("Tile Width: %i   Height: %i\n", theme->tile_width,
		    theme->tile_height);

	    gdk_imlib_destroy_image(img);
	    g_free(full_path);
	}
    }
}

/** 
 * Returns the filename of the image specified through the type of object
 * and the image id.
 * theme - theme reference
 * type - type of object
 * id - the image id
 * Returns: the path to the image file. 
 */
gchar*
theme_get_object_image(Theme* theme, AtomixType type, guint id)
{
    gchar* file = NULL;

    if(theme!=NULL) 
    {
	switch(type)
	{
	case ATOMIX_OBSTACLE:
	    if(theme->obstacles!=NULL) 
	    {
		file = (char*) g_ptr_array_index(theme->obstacles, id);
	    }
	    break;
	    
	case ATOMIX_MOVEABLE:
	    if(theme->moveables)
	    {
		file = (char*)g_ptr_array_index(theme->moveables, id);
	    }
	    break;

	case ATOMIX_DECOR:
	    if(theme->decor)
	    {
		file = (char*)g_ptr_array_index(theme->decor, id);
	    }
	    break;
	    
	default:
	}
    }

    if(file)
    {
	/* add full path */
	file = g_strjoin("/", theme->path, file, NULL);
    }

    return file;
}

/**
 * Sets the absolut path to the themes directory.
 * theme - the theme object
 * path - the path
 */
void
theme_set_path(Theme* theme, gchar* path)
{
    theme->path = g_strdup(path);
}

void
theme_create_hash_table(void)
{
    gint i, n_dirs;
    gchar* users_home;
    gchar* search_dirs[3];
    struct dirent *dent = NULL;
    DIR* dir;

    /* init hash table */
    available_themes = g_hash_table_new((GHashFunc) g_str_hash, 
					(GCompareFunc) g_str_equal);

    /* define directorys where to search for themes */
    users_home = g_get_home_dir();

    n_dirs = 3;
    search_dirs[0] = g_strdup("../themes");
    search_dirs[1] = g_strjoin("/", users_home, ".atomix/themes", NULL);
    search_dirs[2] = g_strjoin("/", DATADIR, "atomix/themes", NULL);
    
    /* search the dirs */
    for(i = 0; i < n_dirs; i++)
    {
	dir = opendir(search_dirs[i]);
	if(dir)
	{
	    char* filename;
	    char* themename;
	    char* directory;
	    while((dent = readdir(dir)) != NULL)
	    {
		if((g_strcasecmp(".", dent->d_name)!=0) &&
		   (g_strcasecmp("..", dent->d_name)!=0))
		{
		    directory = g_strjoin("/", search_dirs[i], dent->d_name,  NULL);
		    filename = g_strjoin("/", directory, "theme", NULL);
		    g_print("Filename: %s\n", filename);

		    themename = theme_get_theme_name_xml(filename);
		    
		    if(themename)
		    {
			g_hash_table_insert(available_themes,
					    g_strdup(themename),
					    g_strdup(directory));
		    }
		    g_free(filename);
		    g_free(directory);
		    g_free(themename);
		}
	    }
	}
#ifdef DEBUG
	else
	{
	    /*  g_print("level_create_hash_table: Opendir error\n"); */
	}
#endif
	g_free(dent);
	closedir(dir);
    }

    for(i = 0; i < n_dirs; i++) g_free(search_dirs[i]);

#ifdef DEBUG
/*    g_print("Scanning for Themes, found: \n");
      g_hash_table_foreach(available_themes, (GHFunc) theme_print_hash_table, NULL); */
#endif
}

void
theme_print_hash_table(gpointer key, gpointer value, gpointer user_data)
{
    g_print("Key: %s, Value: %s\n",(gchar*) key, (gchar*) value);
}

void
theme_destroy_hash_table(void)
{
    if(available_themes)
    {
	g_hash_table_foreach(available_themes, 
			     (GHFunc) theme_destroy_hash_table_item, NULL);
	g_hash_table_destroy(available_themes);
    }
}


void
theme_destroy_hash_table_item(gpointer key, gpointer value, gpointer user_data)
{
    g_free(key);
    g_free(value);
}


/**
 * Loads a theme.
 * name - the name of the theme
 * result - shows if something goes wrong.
 * Returns: pointer to a new created theme object.
 */
Theme*
theme_load(gchar* name, LoadResult* result)
{
    Theme* theme = NULL;
    gint fd;
    gchar* full_path = NULL;
    gchar* dir_path = NULL;

    *result = OK;

    /* setting paths */
    dir_path = (gchar*) g_hash_table_lookup(available_themes, name);
    full_path = g_strjoin("/", dir_path, "theme", NULL);

    /* open file */
    fd = open(full_path, O_RDONLY);

    if(fd==-1) 
    {
	g_print("Theme not found!\n");
	*result = NOT_FOUND;
    }
    else
    {
	/* start parsing */
	theme = theme_new();
	theme_set_path(theme, dir_path);
	if(!theme_parse_theme(fd, theme))
	{
	    g_print("This is not a valid AtomiX theme!\n");
	    *result = NOT_A_VALID_FILE;
	    theme_destroy(theme);
	    theme = NULL;
	}
    }

    if(full_path) g_free(full_path);
    if(dir_path) g_free(dir_path);
    close(fd);

    return theme;    
}


Theme*
theme_load_xml(gchar *name, LoadResult *result)
{
    xmlDocPtr doc;
    xmlNodePtr node;
    Theme *theme;
    gchar *prop_value;
    gchar *dir_path;
    gchar *file_path;

    theme = theme_new();

    /* read file */
    dir_path = (gchar*) g_hash_table_lookup(available_themes, name);
    theme_set_path(theme, dir_path);
    file_path = g_strjoin("/", dir_path, "theme", NULL);
    
    doc = xmlParseFile(file_path); 
       
    if(doc != NULL)
    {
	node = doc->root;
	while(node!=NULL)
	{
	    if(g_strcasecmp(node->name,"THEME")==0)
	    {
		/* handle theme node */
		prop_value = xmlGetProp(node, "name");
		theme->name = g_strdup(prop_value);
		g_print("Theme Name: %s\n", theme->name);
		
		node = node->childs;
	    }
	    else 
	    {	
		if(g_strcasecmp(node->name,"MOVEABLE")==0)
		{
		    /* handle moveable node */
		    theme_fill_img_array(theme, node->childs, ATOMIX_MOVEABLE);
		}
		if(g_strcasecmp(node->name,"OBSTACLE")==0)
		{
		    /* handle obstacle node */
		    theme_fill_img_array(theme, node->childs, ATOMIX_OBSTACLE);
		}
		if(g_strcasecmp(node->name,"DECOR")==0)
		{
		    /* handle decor node */
		    theme_fill_img_array(theme, node->childs, ATOMIX_DECOR);
		}
		if(g_strcasecmp(node->name,"ANIMSTEP")==0)
		{
		    /* handle animstep node */
		    prop_value = xmlGetProp(node, "dist");
		    theme->animstep = atoi(prop_value);
		    g_print("Theme animstep: %s\n", prop_value);
		}
		if(g_strcasecmp(node->name,"BGCOLOR")==0)
		{
		    /* handle background color */
		    prop_value = xmlGetProp(node, "color");
		    theme->bgcolor = g_strdup(prop_value);
		    g_print("bgcolor: %s\n", theme->bgcolor);
		}
		if(g_strcasecmp(node->name,"BGOUTLINE")==0)
		{
		    /* handel background outline color */
		    prop_value = xmlGetProp(node, "color");
		    theme->bgoutline = g_strdup(prop_value);
		    g_print("bgoutline: %s\n", theme->bgoutline);
		}
		
		node = node->next;
	    }
	}
	
	xmlFreeDoc(doc);
    }
    else
    {
	return NULL;
    }

    return theme;
}

void 
theme_fill_img_array(Theme *theme, xmlNodePtr node, AtomixType type)
{
    gchar *img_no;
    gchar *img_src;

    while(node != NULL)
    {
	img_src = xmlGetProp(node, "src");
	theme_add_object_image(theme, type, g_strdup(img_src));
	node = node->next;
    }
}


void
theme_save_xml(Theme *theme, gchar *filename)
{
    xmlDocPtr doc;
    xmlAttrPtr attr;
    xmlNodePtr node;
    xmlNodePtr theme_node;
    xmlNodePtr obstacle_node;
    xmlNodePtr moveable_node;
    xmlNodePtr decor_node;
    xmlNodePtr child_node;
    /* xmlNodePtr decor_node;  not supported yet */
    gint i;
    gint length, dist;
    gchar *str_buffer;
    gchar *img_src;

    str_buffer = g_malloc(5*sizeof(gchar));
    
    /* create xml doc */    
    doc = xmlNewDoc("1.0");
    xmlSetDocCompressMode(doc, 9);

    /* theme name */
    theme_node = xmlNewDocNode(doc, NULL, "THEME", NULL);
    doc->root = theme_node;
    attr = xmlSetProp(theme_node, "name", g_strdup(theme->name));

    /* bg color */
    node = xmlNewChild(theme_node, NULL, "BGCOLOR", NULL);
    attr = xmlSetProp(node, "color", g_strdup(theme->bgcolor));

    /* bg outline color */
    node = xmlNewChild(theme_node, NULL, "BGOUTLINE", NULL);
    attr = xmlSetProp(node, "color", g_strdup(theme->bgoutline));

    /* animstep */
    node = xmlNewChild(theme_node, NULL, "ANIMSTEP", NULL);
    dist = g_snprintf(str_buffer, 5, "%i", theme->animstep);
    attr = xmlSetProp(node, "dist", g_strdup(str_buffer));

    /* obstacle section */
    obstacle_node = xmlNewChild(theme_node, NULL, "OBSTACLE", NULL);
    for(i = 0; i < theme->obstacles->len; i++)
    {
	child_node = xmlNewChild(obstacle_node, NULL, "IMAGE", NULL);
	length = g_snprintf(str_buffer, 5, "%i", i);
	attr = xmlSetProp(child_node, "no", g_strdup(str_buffer));

	img_src = (gchar*)g_ptr_array_index(theme->obstacles, i);
	attr = xmlSetProp(child_node, "src", g_strdup(img_src));
    }
    
    /* moveable section */
    moveable_node = xmlNewChild(theme_node, NULL, "MOVEABLE", NULL);
    for(i = 0; i < theme->moveables->len; i++)
    {
	child_node = xmlNewChild(moveable_node, NULL, "IMAGE", NULL);
	length = g_snprintf(str_buffer, 5, "%i", i);
	attr = xmlSetProp(child_node, "no", g_strdup(str_buffer));

	img_src = (gchar*)g_ptr_array_index(theme->moveables, i);
	attr = xmlSetProp(child_node, "src", g_strdup(img_src));
    }

    /* decor section */
    decor_node = xmlNewChild(theme_node, NULL, "DECOR", NULL);
    for(i = 0; i < theme->decor->len; i++)
    {
	child_node = xmlNewChild(decor_node, NULL, "IMAGE", NULL);
	length = g_snprintf(str_buffer, 5, "%i", i);
	attr = xmlSetProp(child_node, "no", g_strdup(str_buffer));

	img_src = (gchar*)g_ptr_array_index(theme->decor, i);
	attr = xmlSetProp(child_node, "src", g_strdup(img_src));
    }

    xmlSaveFile(filename, doc);

    xmlFreeDoc(doc);
    g_free(str_buffer);
}

gchar*
theme_get_theme_name_xml(gchar *filename)
{
    xmlDocPtr doc;
    xmlNodePtr node;
    gchar* name = NULL;
    gchar *prop_value;

    /* read file */
    doc = xmlParseFile(filename); 
       
    if(doc != NULL)
    {
	node = doc->root;
	
	if((node!=NULL) &&
	   (g_strcasecmp(node->name,"THEME")==0))
	{
	    /* handle level node */
	    prop_value = xmlGetProp(node, "name");
	    name = g_strdup(prop_value);
	}
    }

    xmlFreeDoc(doc);
    
    return name;
}

/** 
 * Retrieves the name of the theme.
 * filename - full qualified path to the theme description.
 * Returns: the name of the theme. 
 */
gchar*
theme_get_theme_name(gchar* filename)
{
    GScanner *scanner;
    gint fd;
    gint i;
    guint token;
    gchar* name  = NULL;

    scanner = g_scanner_new ((GScannerConfig *) &theme_scanner_config);
    fd = open(filename, O_RDONLY);
    
    if(fd==-1)
    {
	/* file not found */
	g_scanner_destroy(scanner);
	return NULL;
    }
    else 
    {

	g_scanner_freeze_symbol_table (scanner);
	for (i = 0; i < n_symbols; i++)
	{
	    g_scanner_add_symbol (scanner, symbols[i].name, 
				  GINT_TO_POINTER (symbols[i].token));
	}
	g_scanner_thaw_symbol_table (scanner);

	g_scanner_input_file(scanner, fd);

	if (g_scanner_peek_next_token(scanner) != G_TOKEN_EOF)
	{
	    token = g_scanner_get_next_token(scanner);
	    if((token == THEME_TOKEN_THEME) &&
	       (g_scanner_peek_next_token(scanner) != G_TOKEN_EOF))
	    {
		token = g_scanner_get_next_token(scanner);
		
		if(token == G_TOKEN_STRING)
		{
		    name = g_strdup(scanner->value.v_string); 
		} 
		else g_print("No Identifier\n");
	    } 
	    else g_print("No Theme token\n");	    
	}
	else g_print("EOF\n");	
    }
    g_scanner_destroy(scanner);
    close(fd);

    return name;
}


/**
 * Main theme parse function.
 * fd - file descriptor of the theme file.
 * theme - pointer to a theme object which will be filled 
 *         with the appropriate data
 * Returns: FALSE if an parse error occured, else TRUE
 */
gboolean
theme_parse_theme(guint fd, Theme* theme)
{
    GScanner *scanner;
    int i;
    ThemeFSM state = THEME_NONE_READ;
    guint token;
    gboolean parse_error = FALSE;

    scanner = g_scanner_new ((GScannerConfig *) &theme_scanner_config);
    g_scanner_input_file(scanner, fd);
    
    g_scanner_freeze_symbol_table (scanner);
    for (i = 0; i < n_symbols; i++)
    {
	g_scanner_add_symbol (scanner, symbols[i].name, 
			      GINT_TO_POINTER (symbols[i].token));
    }
    g_scanner_thaw_symbol_table (scanner);

    while((g_scanner_peek_next_token(scanner) != G_TOKEN_EOF) &&
	  !parse_error && (state != THEME_CLOSED_BRACE))
    {
	token = g_scanner_get_next_token(scanner);
	switch(state)
	{
	case THEME_NONE_READ:
	    if(token == THEME_TOKEN_THEME)
	    {
		state = THEME_THEME_READ;
	    }
	    else
	    {
		parse_error = TRUE;
	    }
	    break;

	case THEME_THEME_READ:
	    if(token == G_TOKEN_STRING)
	    {
		theme->name = g_strdup(scanner->value.v_string);
		state = THEME_IDENTIFIER_READ;
	    }
	    else
	    {
		parse_error = TRUE;
	    }
	    break;

	case THEME_IDENTIFIER_READ:
	    if(token == G_TOKEN_LEFT_CURLY)
	    {
		state = THEME_CONTENT;
	    }
	    else
	    {
		parse_error = TRUE;
	    }
	    break;

	case THEME_CONTENT:
	    switch(token)
	    {
	    case THEME_TOKEN_MOVEABLE:
		parse_error = theme_parse_moveable(scanner, theme);
		break;

	    case THEME_TOKEN_OBSTACLE:
		parse_error = theme_parse_obstacle(scanner, theme);
		break;

	    case THEME_TOKEN_DECOR:
		parse_error = theme_parse_decor(scanner, theme);
		break;

	    case G_TOKEN_RIGHT_CURLY:
		state = THEME_CLOSED_BRACE;
		break;

	    default:
		parse_error = TRUE;
	    }
	    break;

	default:
	    parse_error = TRUE;
	}
    }

    g_scanner_destroy(scanner);

    if(state!=THEME_CLOSED_BRACE) parse_error = TRUE;

    return !parse_error;
}

/**
 * Parse the MOVEABLE section of a theme description.
 * scanner - gnome scanner object
 * theme - to be filled theme object
 * Returns: TRUE, if a parse error occured, else FALSE
 */
gboolean
theme_parse_moveable(GScanner* scanner, Theme* theme)
{
    return theme_parse_image_list(scanner, theme, ATOMIX_MOVEABLE);
}

/**
 * Parse the OBSTACLE section of a theme description.
 * scanner - gnome scanner object
 * theme - to be filled theme object
 * Returns: TRUE, if a parse error occured, else FALSE
 */
gboolean
theme_parse_obstacle(GScanner* scanner, Theme* theme)
{
    return theme_parse_image_list(scanner, theme, ATOMIX_OBSTACLE);
}

/**
 * Parse the DECOR section of a theme description.
 * scanner - gnome scanner object
 * theme - to be filled theme object
 * Returns: TRUE, if a parse error occured, else FALSE
 */
gboolean
theme_parse_decor(GScanner* scanner, Theme* theme)
{
    return theme_parse_image_list(scanner, theme, ATOMIX_DECOR);
}

/**
 * Generic parse function for the image lists. This will be called 
 * by the previous functions.
 * scanner - gnome scanner object
 * theme - to be filled theme object
 * type - the type of object to which the images will be added. 
 * Returns: TRUE, if a parse error occured, else FALSE
 */
gboolean
theme_parse_image_list(GScanner* scanner, Theme* theme, AtomixType type)
{
    gboolean parse_error = FALSE;
    ThemeFSM state = THEME_NONE_READ;
    guint token = G_TOKEN_NONE;
    
    while((g_scanner_peek_next_token(scanner) != G_TOKEN_EOF) &&
	  !parse_error && (state != THEME_CLOSED_BRACE))
    {
	token = g_scanner_get_next_token(scanner);
	switch(state)
	{
	case THEME_NONE_READ:
	    //g_print("THEME_NONE_READ\n");
	    if(token==G_TOKEN_EQUAL_SIGN)
	    {
		state = THEME_EQUAL_READ;
	    }
	    else
	    {
		parse_error = TRUE;
	    }	    
	    break;

	case THEME_EQUAL_READ:
	    //g_print("THEME_EQUAL_READ\n");
	    if(token==G_TOKEN_LEFT_CURLY)
	    {
		state = THEME_OPEN_BRACE;
	    }
	    else
	    {
		parse_error = TRUE;
	    }	    
	    break;

	case THEME_OPEN_BRACE:
	    //g_print("THEME_OPEN_BRACE\n");
	    if(token == G_TOKEN_INT)
	    {
		state = THEME_NUMBER_READ;
	    }
	    else if(token == G_TOKEN_STRING)
	    {		
		theme_add_object_image(theme, type, g_strdup(scanner->value.v_string));
		state = THEME_IMAGE_READ;
	    }
	    else if(token == G_TOKEN_RIGHT_CURLY)
	    {
		state = THEME_CLOSED_BRACE;
	    }
	    else 
	    {
		parse_error = TRUE;
	    }	    
	    break;

	case THEME_NUMBER_READ:
	    //g_print("THEME_NUMBER_READ\n");
	    if(token == G_TOKEN_EQUAL_SIGN)
	    {
		state = THEME_NUMBER_EQUAL_READ;
	    }
	    else 
	    {
		parse_error = TRUE;
	    }	    
	    break;	    

	case THEME_NUMBER_EQUAL_READ:
	    //g_print("THEME_NUMBER_EQUAL_READ\n");
	    if(token == G_TOKEN_STRING)
	    {		
		theme_add_object_image(theme, type, g_strdup(scanner->value.v_string));
		state = THEME_IMAGE_READ;
	    }
	    else 
	    {
		parse_error = TRUE;
	    }	    
	    break;	    

	case THEME_IMAGE_READ:
	    //g_print("THEME_IMAGE_READ\n");
	    if(token == G_TOKEN_COMMA)
	    {		
		state = THEME_OPEN_BRACE;
	    }
	    else if(token == G_TOKEN_RIGHT_CURLY)
	    {
		state = THEME_CLOSED_BRACE;
	    }
	    else 
	    {
		parse_error = TRUE;
	    }	    
	    break;	    	    

	default:
	    parse_error = TRUE;
	}
    }

    if(state!=THEME_CLOSED_BRACE) parse_error = TRUE;
    return parse_error;
}

void
theme_tile_size(Theme *theme, gint *width, gint *height)
{
   *width = theme->tile_width;
   *height = theme->tile_height;
}


